/* Gnome Scan - Scan as easy as you print
 * Copyright © 2007  Étienne Bersac <bersace03@laposte.net>
 *
 * Gnome Scan is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * gnome-scan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gnome-scan.  If not, write to:
 *
 *	the Free Software Foundation, Inc.
 *	51 Franklin Street, Fifth Floor
 *	Boston, MA 02110-1301, USA
 */

/**
 * SECTION: gnome-scan-settings
 * @short_description: Store settings for plugins
 *
 * When user switch from a scanner to another, he expects the settings
 * to keep almost the same. e.g. resolution should not be resetted on
 * device selection. The settings are remember in a #GnomeScanSettings
 * which maintains a table of #GValue using @GParamSpec:name as
 * key. This allow to apply the same #GnomeScanSettings on various
 * #GnomeScanPlugin through gnome_scan_plugin_configure().
 *
 * Since 0.5.92, #GnomeScanSettings load and store automatically
 * settings in #GConf in /apps/gnome-scan/%prgname%/%key%. Ensure you
 * have a call to g_set_prgname() before gnome_scan_init(). Loading is
 * done upon construction. However, storing is done at dispose
 * stage. Ensure to correctly ref/unref your #GnomeScanSettings
 * instance in order to actually use this feature.
 **/

#include <glib/gi18n.h>
#include <gconf/gconf-client.h>
#include "gnome-scan-settings.h"

#define	GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GNOME_TYPE_SCAN_SETTINGS, GnomeScanSettingsPrivate))

typedef struct _GnomeScanSettingsPrivate GnomeScanSettingsPrivate;

struct _GnomeScanSettingsPrivate
{
	/* using GParamSpec name as key, and GValue as value */
	GHashTable	*values;
	GConfClient	*gconf;
	gboolean	disposed;
};

enum
  {
    CHANGED,
    N_SIGNALS
  };

static GObjectClass* parent_class = NULL;
static guint signals[N_SIGNALS] = {0};

G_DEFINE_TYPE (GnomeScanSettings, gnome_scan_settings, G_TYPE_OBJECT);

static void
gnome_scan_settings_init (GnomeScanSettings *object)
{
	GnomeScanSettingsPrivate *priv = GET_PRIVATE (object);
	
	priv->values = g_hash_table_new (g_str_hash, g_str_equal);
	priv->gconf = gconf_client_get_default();
}

static gchar*
gss_gconf_key(gchar *name)
{
	gchar *key;
	gchar *tmp = g_strdup_printf ("/apps/gnome-scan/%s", g_get_prgname ());
	
	if (name)
		key = g_strdup_printf("%s/%s", tmp, name);
	else
		key = g_strdup(tmp);
		
	g_free(tmp);
	
	return key;
}

static GValue*
gss_gconf_value_to_gvalue(GConfValue *value)
{
	GType gtype = G_TYPE_NONE;
	GValue *gvalue = g_new0(GValue, 1);
	
	switch(value->type) {
		case GCONF_VALUE_STRING:
			gtype = G_TYPE_STRING;
			break;
		case GCONF_VALUE_INT:
			gtype = G_TYPE_INT;
			break;
		case GCONF_VALUE_FLOAT:
			gtype = G_TYPE_DOUBLE;
			break;
		case GCONF_VALUE_BOOL:
			gtype = G_TYPE_BOOLEAN;
			break;
	}
	
	g_value_init(gvalue, gtype);
	
	switch(value->type) {
		case GCONF_VALUE_STRING:
			g_value_set_static_string(gvalue,
									  gconf_value_get_string(value));
			break;
		case GCONF_VALUE_INT:
			g_value_set_int(gvalue,
							gconf_value_get_int(value));
			break;
		case GCONF_VALUE_FLOAT:
			g_value_set_double(gvalue,
							  gconf_value_get_float(value));
			break;
		case GCONF_VALUE_BOOL:
			g_value_set_boolean(gvalue,
								gconf_value_get_bool(value));
			break;
	}
	
	if (gtype == G_TYPE_NONE) {
		g_free(gvalue);
		gvalue = NULL;
	}
	
	return gvalue;
}

static GConfValue*
gss_gvalue_to_gconf_value(GValue *gvalue)
{
	GConfValueType type = GCONF_VALUE_INVALID;
	GConfValue *value;
	GValue *trans = g_new0(GValue, 1);
	GType gtype = G_VALUE_TYPE(gvalue);
	switch(gtype) {
		case G_TYPE_STRING:
			type = GCONF_VALUE_STRING;
			break;
		case G_TYPE_INT:
			type = GCONF_VALUE_INT;
			break;
		case G_TYPE_FLOAT:
		case G_TYPE_DOUBLE:
			type = GCONF_VALUE_FLOAT;
			break;
		case G_TYPE_BOOLEAN:
			type = GCONF_VALUE_BOOL;
			break;
		default:
			if (g_type_is_a (gtype, G_TYPE_ENUM)) {
				type = GCONF_VALUE_INT;
				gtype = G_TYPE_INT;
			}
			else
				return NULL;
	}
	
	value = gconf_value_new(type);
	
	switch(gtype) {
		case G_TYPE_STRING:
			gconf_value_set_string(value,
								   g_value_get_string(gvalue));
			break;
		case G_TYPE_INT:
			g_value_init(trans, G_TYPE_INT);
			g_value_transform (gvalue, trans);
			gconf_value_set_int(value,
								g_value_get_int(trans));
			break;
		case G_TYPE_FLOAT:
		case G_TYPE_DOUBLE:
			g_value_init(trans, G_TYPE_FLOAT);
			g_value_transform (gvalue, trans);
			gconf_value_set_float(value,
								g_value_get_float(trans));
			break;
		case G_TYPE_BOOLEAN:
			gconf_value_set_bool(value,
								 g_value_get_boolean (gvalue));
			break;
	}
	
	if (!GCONF_VALUE_TYPE_VALID(type)) {
		g_free(value);
		value = NULL;
	}
	
	g_free(trans);
	
	return value;
}

static GObject*
gnome_scan_settings_constructor (GType type,
								 guint n_params,
								 GObjectConstructParam *params)
{
	GObject *object;
	GnomeScanSettingsPrivate *priv;
	GSList *entries;
	gchar *dir = gss_gconf_key(NULL);
	gchar *name;
	GError *error = NULL;
	GConfEntry *entry;
	GValue *gvalue;
	
	object =
		G_OBJECT_CLASS(parent_class)->constructor(type,
												  n_params,
												  params);
	priv = GET_PRIVATE(object);
	entries = gconf_client_all_entries(priv->gconf, dir, &error);
	for (; entries; entries = entries->next) {
		entry = entries->data;
		name = g_path_get_basename(entry->key);
		if (gvalue = gss_gconf_value_to_gvalue(entry->value))
				gnome_scan_settings_set(GNOME_SCAN_SETTINGS(object),
										name, gvalue);
		g_free(name);
	}
	
	g_free(dir);
	
	return object;
}

static void
gnome_scan_settings_finalize (GObject *object)
{
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

void
gss_store_key(gchar *name, GValue *gvalue, GnomeScanSettings *settings)
{
	GnomeScanSettingsPrivate *priv = GET_PRIVATE(settings);
	GConfValue *value;
	gchar *key = gss_gconf_key(name);
	GError *error = NULL;
	
	if (value = gss_gvalue_to_gconf_value(gvalue)) {
		gconf_client_set(priv->gconf,
						 key, value,
						 &error);

		if (error) {
			g_warning("Failed saving %s option : %s",
					  name, dgettext(g_quark_to_string (error->domain),
									 error->message));
		}
	}

	g_free(key);
}

static void
gnome_scan_settings_dispose (GObject *object)
{
	GnomeScanSettingsPrivate *priv = GET_PRIVATE(object);
	
	if (priv->disposed == FALSE) {
		g_hash_table_foreach (priv->values,
							  (GHFunc) gss_store_key,
							  object);
		priv->disposed = TRUE;
	}
	
	G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
gnome_scan_settings_class_init (GnomeScanSettingsClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	
	g_type_class_add_private (klass, sizeof (GnomeScanSettingsPrivate));
	object_class->constructor   = gnome_scan_settings_constructor;
	object_class->finalize		= gnome_scan_settings_finalize;
	object_class->dispose		= gnome_scan_settings_dispose;
	
  /**
   * GnomeScanSettings::changed:
   * @widget:	The emitting widget
   * @key:	The changed key
   * @value:	The new value
   *
   * This signal is emitted when the value is changed by the user.
   **/
  signals[CHANGED] =
    g_signal_new ("changed",
		  G_OBJECT_CLASS_TYPE (klass),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (GnomeScanSettingsClass, changed),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__STRING,
		  G_TYPE_NONE, 1, G_TYPE_STRING);
}




/**
 * gnome_scan_settings_new:
 * 
 * Create a new empty #GnomeScanSettings.
 * 
 * Returns: a new #GnomeScanSettings
 **/
GnomeScanSettings*
gnome_scan_settings_new (void)
{
  GObject *object = g_object_new (GNOME_TYPE_SCAN_SETTINGS, NULL);
  return GNOME_SCAN_SETTINGS (object);
}

/**
 * gnome_scan_settings_init_value:
 * @settings:	a #GnomeScanSettings
 * @pspec:	a #GSParamSpec
 *
 * Store a new value in @settings using default value and key from
 * @pspec.
 **/
void
gnome_scan_settings_init_value (GnomeScanSettings *settings,
								GParamSpec *pspec)
{
	/* TODO: store schemas in gconf */
	g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
	GValue *value = g_new0 (GValue, 1);
	g_value_init (value, G_PARAM_SPEC_VALUE_TYPE (pspec));
	g_param_value_set_default (pspec, value);
	gnome_scan_settings_set (settings,
							 g_param_spec_get_name (pspec),
							 value);
	g_value_reset (value);
	g_free (value);
}

/**
 * gnome_scan_settings_set:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: #GParamSpec user value
 * 
 * Store @value for @name key.
 * See: gnome_scan_settings_get()
 **/
void
gnome_scan_settings_set (GnomeScanSettings *settings,
                         const gchar *name,
                         const GValue *value)
{
	g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings) && G_IS_VALUE (value));
	g_hash_table_insert (GET_PRIVATE (settings)->values, g_strdup (name),
						 g_boxed_copy (G_TYPE_VALUE, value));
	g_signal_emit (settings, signals[CHANGED], 0, name);
}

/**
 * gnome_scan_settings_set_transform:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: #GParamSpec user value
 * 
 * Store @value for @name key, keeping existing value type by
 * transforming @value. @value type is used if the @name setting has
 * never been set.
 *
 * See: gnome_scan_settings_get_transformed()
 **/
void
gnome_scan_settings_set_transform (GnomeScanSettings *settings,
								   const gchar *name,
								   const GValue *value)
{
	g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings) && G_IS_VALUE (value));
	GValue *v = gnome_scan_settings_get (settings, name);
	if (v) {
		g_value_transform (value, v);
	}
	else {
		v = g_boxed_copy (G_TYPE_VALUE, value);
	}
	
	g_hash_table_insert (GET_PRIVATE (settings)->values, g_strdup (name),
						 v);
	g_signal_emit (settings, signals[CHANGED], 0, name);
}

/**
 * gnome_scan_settings_set_double:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: gdouble value
 * 
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_double (GnomeScanSettings *settings,
                                const gchar *name,
                                gdouble value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, G_TYPE_DOUBLE);
  g_value_set_double (v, value);
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_set_int:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: gint value
 * 
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_int (GnomeScanSettings *settings,
                             const gchar *name,
                             gint value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, G_TYPE_INT);
  g_value_set_int (v, value);
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_set_boolean:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: gboolean value
 * 
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_boolean (GnomeScanSettings *settings,
                             const gchar *name,
                             gboolean value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, G_TYPE_BOOLEAN);
  g_value_set_boolean (v, value);
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_set_enum:
 * @settings:	a #GnomeScanSettings
 * @name:	#GParamSpec name
 * @type:	enum GType
 * @value:	enum value
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_enum (GnomeScanSettings *settings,
                              const gchar *name,
                              GType type,
                              gint value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, type);
  g_value_set_enum (v, value);
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_set_boxed:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: a #GBoxed
 * 
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_boxed (GnomeScanSettings *settings,
                               const gchar *name,
                               GType type,
                               gpointer value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, type);
  g_value_set_boxed (v, g_boxed_copy (type, value));
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_set_object:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: a #GObject
 * 
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_object (GnomeScanSettings *settings,
                                const gchar *name,
                                GObject* value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, G_TYPE_OBJECT);
  g_value_set_object (v, value);
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_set_pointer:
 * @settings: a #GnomeScanSettings
 * @name: #GParamSpec name
 * @value: a gpointer
 * 
 *
 * See: gnome_scan_settings_set_transform()
 **/
void
gnome_scan_settings_set_pointer (GnomeScanSettings *settings,
                                 const gchar *name,
                                 gpointer value)
{
  g_return_if_fail (GNOME_IS_SCAN_SETTINGS (settings));
  GValue *v = g_new0 (GValue, 1);
  g_value_init (v, G_TYPE_POINTER);
  g_value_set_pointer (v, value);
  gnome_scan_settings_set_transform (settings, name, v);
}

/**
 * gnome_scan_settings_get:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the value corresponding to @key or NULL.
 * See: gnome_scan_settings_set()
 **/
GValue*
gnome_scan_settings_get (GnomeScanSettings *settings,
                         const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), NULL);
  GValue *v = g_hash_table_lookup (GET_PRIVATE (settings)->values, name);
  if (v)
    v = g_boxed_copy (G_TYPE_VALUE, v);
	
  return v;
}

/**
 * gnome_scan_settings_get_transformed:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * @type:	Transformation destination type.
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the value corresponding to @key or NULL.
 * See: gnome_scan_settings_set()
 **/
GValue*
gnome_scan_settings_get_transformed	(GnomeScanSettings *settings,
                                        const gchar *name,
                                        GType type)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), NULL);
  GValue *v = gnome_scan_settings_get (settings, name);
  GValue *t = g_new0 (GValue, 1);
  g_value_init (t, type);
  if (v) {
    g_value_transform (v, t);
    g_value_unset (v);
    g_free (v);
  }
  else {
    g_warning ("Option %s not set", name);
  }
  return t;
}


/**
 * gnome_scan_settings_get_string:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the string corresponding to @key.
 * See: gnome_scan_settings_get()
 **/
gchar*
gnome_scan_settings_get_string (GnomeScanSettings *settings,
                                const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), NULL);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   G_TYPE_STRING);
  gchar* val = g_value_dup_string (v);
    g_value_unset(v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_double:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the string corresponding to @key or NULL.
 * See: gnome_scan_settings_get()
 **/
gdouble
gnome_scan_settings_get_double (GnomeScanSettings *settings,
                                const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), 0.);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   G_TYPE_DOUBLE);
  gdouble val = g_value_get_double (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_int:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the string corresponding to @key or NULL.
 * See: gnome_scan_settings_get()
 **/
gint
gnome_scan_settings_get_int (GnomeScanSettings *settings,
                             const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), 0);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   G_TYPE_INT);
  gint val = g_value_get_int (v);
  g_value_unset (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_boolean:
 * @settings:	a #GnomeScanSettings
 * @name:	value name
 *
 * Returns: a gboolean 
 **/
gboolean
gnome_scan_settings_get_boolean (GnomeScanSettings *settings,
                             const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), 0);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,  G_TYPE_BOOLEAN);
  gboolean val = g_value_get_boolean (v);
  g_value_unset (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_enum:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * @type: #GEnum subtype
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the enum value corresponding to @key or 0.
 * See: gnome_scan_settings_get()
 **/
gint
gnome_scan_settings_get_enum (GnomeScanSettings *settings,
			      const gchar *name,
			      GType type)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), 0);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   type);
  gint val = g_value_get_enum (v);
  g_value_unset (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_boxed:
 * @settings: a #GnomeScanSettings
 * @type:	#GBoxed derived #GType
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the boxed-type value corresponding to @key or NULL.
 * See: gnome_scan_settings_get()
 **/
gpointer
gnome_scan_settings_get_boxed (GnomeScanSettings *settings,
			       const gchar *name,
			       GType type)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), NULL);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   type);
  gpointer val = g_value_dup_boxed (v);
  g_value_unset (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_object:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the object corresponding to @key or NULL.
 * See: gnome_scan_settings_get()
 **/
GObject*
gnome_scan_settings_get_object (GnomeScanSettings *settings,
				const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), NULL);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   G_TYPE_OBJECT);
  GObject* val = g_value_dup_object (v);
  g_value_unset (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_pointer:
 * @settings: a #GnomeScanSettings
 * @name: a key
 * 
 * Look for a value stored in the table for @name key.
 * 
 * Returns: the pointer corresponding to @key or NULL.
 * See: gnome_scan_settings_get()
 **/
gpointer
gnome_scan_settings_get_pointer (GnomeScanSettings *settings,
				 const gchar *name)
{
  g_return_val_if_fail (GNOME_IS_SCAN_SETTINGS (settings), NULL);
  GValue *v = gnome_scan_settings_get_transformed (settings,
						   name,
						   G_TYPE_POINTER);
  gpointer val = g_value_get_pointer (v);
  g_value_unset (v);
  g_free (v);
  return val;
}

/**
 * gnome_scan_settings_get_all:
 * @settings: a #GnomeScanSettings
 *
 * Retrun all stored values in a #GHashTable. Use g_hash_table_ref() if 
 * you want to keep the #GHashTable in memory after @settings destruction.
 *
 * Returns: a GHashTable containings #GValues
 */
GHashTable*
gnome_scan_settings_get_all (GnomeScanSettings *settings)
{
  return GET_PRIVATE (settings)->values;
}
